!iCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CCC
!CCC Name       : SUBROUTINE CALRATE 
!CCC Purpose    : To compute thermal rates (or yields) for each product channel
!CCC              as functions of both T and P at a particular angular momentum
!CCC 
!CCC Authors    : Thanh Lam Nguyen (thanhlamnguyen08@gmail.com)
!CCC              John F. Stanton  (jfstanton@gmail.com)
!CCC
!CCC Date       : Apr. 18, 2014
!CCC Where      : The University of Texas at Austin
!CCC
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


       SUBROUTINE CALRATE(Jmin,Jmax,MAXJ, NJ,DELE,Maxi,FE,NN,NTS,Np,IND,Edown,T,Nsink,suJ,rKp,RKc,&       
                ESOLVER, BB, BBT, TSS, TDS, CNAME, CNM, CNP, IL, KL, &
                FEDISS, NDISS, NDI, NDJ, CDSTS, SYMRE, SYMTS )

        IMPLICIT NONE

!!!  Jmin, Jmax  : Input angular momentum number; Jmin = Jmax in this case. 
!!!  MAXJ        : maximum angular momentum number
!!!  NJ          : step size of angular momentum number, J
!!!  DELE        : energy bin (grain energy)
!!!  Maxi        : maximum bin in energy-grained master equation (EGME)
!!!  FE          : initial energy distribution function, F(E,J)
!!!  NN          : a number of minima
!!!  NTS         : a number of TSs
!!!  Np          : a number of products
!!!  IND         : index of the lowest-lying intermediate / adduct
!!!  Edown       : down energy transfer (cm-1)
!!!  T           : temperature (K)
!!!  Nsink       : a vector of sinks that is set to collect stabilized
!!!                populations of intermediates 
!!!  suJ         : sum of F(E) at a particular J    
!!!  rKp         : collision frequency (1/s) at a considered pressure
!!!  RKc         : overall bi-molecular reaction thermal rate (cm3/molecule/s)
!!!                or set RKc = 1 to gain product yields. 
!!!  ESOLVER     : either "DENSM" or "SPASM"
!!!  BB          : vector of three rot. constants for minima
!!!  BBT         : vector of three rot. constants for TSs
!!!  TSS         : number of states for TSs
!!!  TDS         : number of states for minima
!!!  CNAME       : names of stationary points
!!!  CNM         : names of minima
!!!  CNP         : names of products
!!!  IL          : index of reactant
!!!  KL          : index of product(s)
!!!  Sigf        : forward reaction path degeneracy
!!!  Sigr        : reverse reaction path degeneracy
!!!  FEDISS      : "YES" or "NO" : to print out F(E)J at the moment of
!!!                     dissociation
!!!  NDISS       : number of F(E)J functions need to be printed:
!!!                from NDI leading to NDJ
!!!  NDI         : index of reactant/adduct
!!!  NDJ         : index of products

        Include 'declare.inc'

        Integer MAXJ, NJ
        Integer N, M, Np, NTS, IND
        Integer  I, J, K, L, Maxi, NN, NE, JJT, KK, LL
        Integer Jmin, Jmax
        Integer  Nsink(No), ip(No), IL(No), KL(No)

        Double precision SYMRE(No), SYMTS(No)

        Double precision TSS(NTS,Maxi), TDS(NN,Maxi)
        Double precision, Allocatable :: SS(:,:), DS(:,:)

        Double precision, Allocatable :: A(:,:) 
        Double precision, Allocatable :: rktp(:,:,:)
        Double precision, Allocatable :: rk(:,:,:)
        Double precision, Allocatable :: rp(:,:,:)
        Double precision, Allocatable :: P(:,:,:)

        Double precision FE(Maxi,0:MAXJ)

        Double precision b(NN*Maxi), SST(NN*Maxi), tp
        Double precision Rate(NN+Np), RateP(NN+Np)
        Double precision FF(Maxi), Fint(Maxi), RRKM(Maxi)

!!!        Double precision Sigf(No), Sigr(No)
        Double precision S(NN,Maxi)

        Double precision T1(NMAX), AT0(NMAX)
        Double precision SS1(NMAX), DS1(NMAX)

        Double precision PP, beta, T, rKp, RKc
        Double precision E0, Edown, DELE
        Double precision suJJ, su, suJ, suk
        Double precision su2, su3, su4, ENG
        Double precision BT(3)

        Double precision Rate0(NN,NN+Np)

        INTEGER IPIV(NN*Maxi), LDA, LDB, INFO
        INTEGER N1, M1, IT

        CHARACTER*20 CNAME(NM,No)
        CHARACTER*20 CNM(No), CNP(No)
        CHARACTER*5 ESOLVER
        CHARACTER*3 FEDISS
        Integer NDISS, NDI(NM), NDJ(No)
        CHARACTER*12 CDSTS(No)


        EXTERNAL dgesv

        ALLOCATE ( SS(NTS,Maxi), DS(NN,Maxi) )

        DO I=1, Maxi
          AT0(I)=0.0d0
        ENDDO

        beta = 1.0d0/RT/T
 
        IF(Jmin.EQ.Jmax) THEN
                JJT = Jmin
                su=0.0D0
                Do I=1,Maxi
                        su=su+FE(I,JJT)
                Enddo
                suJJ = su
        ELSE
                suJJ = suJ
        ENDIF

        IF(suJJ.eq.0.0d0) GOTO 1111

!CCCCCCCCCCCCCCCCCCCCCCCCCCCC

        DO I=1, NN
          DO J=1, NN + Np 
           Rate0(I,J)=0.0d0   !!! initial reaction rates for each channel
          Enddo
        ENDDO

!!!       IF(suJJ.eq.0.0d0) goto 450   !!! Do not compute when sum of F(E)J = 0

!!!        DO JJT=Jmin, Jmax, NJ  !!! To start J loop

        JJT = Jmin  !!! Note that Jmin = Jmax


!!CCCCCCCCCCCCCCCCCCCCCCCCC

!!! For minima

        DO K=1, NN
          DO I=1, Maxi
                T1(I) = TDS(K,I)
          ENDDO
          DO I=1, 3
                BT(I) = BB(K,I)
          ENDDO
          CALL cdsj(SS1,DS1,T1,AT0,DELE,Maxi,BT,JJT) !!! To convolute E,J
          DO I=1, Maxi
                DS(K,I)=DS1(I)/SYMRE(K)
          ENDDO
        ENDDO
!!CCCCCCCCCCCCCCCCCCCCCCCC

!!! For TSs

        DO K=1, NTS

        IF(CDSTS(K).NE.'VRRKM_HO') THEN

        DO I=1, Maxi
                T1(I) = TSS(K,I)
        ENDDO
        DO I=1, 3
                BT(I) = BBT(K,I)
        ENDDO
        CALL cdsj(SS1,DS1,T1,AT0,DELE,Maxi,BT,JJT)  !!! To convolute E,J
        DO I=1, Maxi
                SS(K,I)=SS1(I)/SYMTS(K)
        ENDDO

        ELSE
                SS(K,1) = TSS(K,1)/SYMTS(K)
                DO J=2, Maxi
                        SS(K,J) = SS(K,J-1) + TSS(K,J)/SYMTS(K)
                ENDDO
        ENDIF
        ENDDO
!!CCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE ( rk(NN,NN+Np,Maxi) )


        DO I=1, NN
        DO J=1, NN + Np
                DO K=1, Maxi
                        rk(I,J,K)=0.0d0  !!! Initial microcanonical rates for
!!!                                             each reaction channel
                ENDDO
        ENDDO
        ENDDO                

        DO L=1, NTS
        IF((IL(L).GT.0).and.(KL(L).GT.0)) THEN
            DO J=1, Maxi
                IF(DS(IL(L),J).GT.0.0d0) THEN                
                  rk(IL(L),KL(L),J) = rk(IL(L),KL(L),J) + SS(L,J)/DS(IL(L),J)/hh
                ENDIF
                IF(KL(L).LE.NN) THEN
                  IF(DS(KL(L),J).GT.0.0d0) THEN 
                  rk(KL(L),IL(L),J) = rk(KL(L),IL(L),J) + SS(L,J)/DS(KL(L),J)/hh
                  ENDIF
                ENDIF
            ENDDO
        ENDIF
        ENDDO

!!CCCCCCCCCCCCCCCCCCCCCC
!!! For an empty grain
!!CCCCCCCCCCCCCCCCCCCCCC

        DO K=1, Maxi 
        DO I=1, NN 
        DO J=1, NN 
                tp = rk(I,J,K)*rk(J,I,K)
                IF(tp.eq.0.0d0) THEN
                        rk(I,J,K)=0.0d0
                        rk(J,I,K)=0.0d0
                ENDIF
        ENDDO
        ENDDO
        ENDDO

!!! END of an empty grain

        IF(DS(IND,1).eq.0.0d0) DS(IND,1)=1.0E-30

        DO I=1, Maxi
                DS1(I)=DS(IND,I)
        ENDDO

        ALLOCATE (P(NN,Maxi,Maxi)) 

	call CalP(P,DELE,Edown,Maxi,T,Nsink,NN,DS1,IND) !!! to compute energy transfer
!!!                                            probability function, P(E)J

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C
!C To compute P(E) and construct a B-vector (|B>) in the equation A|X> = |B> 
!C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	suk=0.0d0
	DO I=1, Maxi
           b(I)=FE(I,JJT)/suJJ
           suk = suk + FE(I,JJT)/suJJ
	ENDDO

	DO KK=2, NN
		DO I=1, Maxi
			b(I+Maxi*(KK-1))=0.0D0
		ENDDO
	ENDDO

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C
!C To input all product channels from each WELL	
!C	
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE (rp(NN,NN+Np,Maxi))

        DO I=1, NN
                ip(I) = Np + 1 
        ENDDO

        DO I=1, NN
                DO J=1, Np + 1 
                IF(J.EQ.1) THEN
                        DO K=1, Maxi 
                                rp(I,J,K)=rKp
                        ENDDO
                ELSE
                        DO K=1, Maxi
                                rp(I,J,K)=rk(I,J+NN-1,K)
                        ENDDO
                ENDIF
                ENDDO
        ENDDO

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CC	Input rate coefficients in kinetic scheme
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE (rktp(NN,NN+Np,Maxi))

        DO K=1, Maxi
        DO I=1, NN
                su = 0.0d0
                DO J=1, Np + NN
                IF(I.ne.J) THEN
                        su = su + rk(I,J,K)
                ENDIF
                ENDDO
                rktp(I,I,K) = su
        ENDDO
        ENDDO

        DO K=1, Maxi
        DO I=1, NN
                DO J=1, NN + Np 
                IF(I.ne.J) THEN
                        rktp(I,J,K) = rk(I,J,K)
                ENDIF
                ENDDO
        ENDDO
        ENDDO

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!CC	Computing a A-matrix
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        ALLOCATE (A(Maxi*NN,Maxi*NN))

	DO KK=1, NN
		DO LL=1, NN
			CALL CalA(A,rKp,P,rktp,Maxi,KK,LL,NN,Np)
		ENDDO
	ENDDO
        
        DEALLOCATE (rktp, P)

	
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
!C
!C To solve AX=B using LAPACK or Iterative SOLVER
!C
!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

       DO I=1, Maxi*NN
          SST(I) = b(I)
       ENDDO

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!! Select a linear equation solver: 
!! dens matrix (DENSM) or sparse matrix (SPASM)
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        IF(ESOLVER.EQ.'DENSM') THEN

        write(*,*) ""
        write(*,*) "==================================="
        write(*,*) "DENSE MATRIX: LAPACK to be used "
        write(*,*) "==================================="
        write(*,*) ""

        LDA = Maxi*NN
        CALL dgesv(Maxi*NN, 1, A, LDA, IPIV, SST, Maxi*NN, INFO)
        IF(INFO.ne.0) THEN
                write(*,*) "Sth is wrong !!!" 
                STOP
        ELSE

        write(*,*) ""
        write(*,*) "==================================="
        write(*,*) "The MISSION of LAPACK is COMPLETE"
        write(*,*) "==================================="
        write(*,*) ""        

        ENDIF

        ELSE IF(ESOLVER.EQ.'SPASM') THEN

        write(*,*) ""
        write(*,*) "=================================="
        write(*,*) "BEGIN to use an Iterative Solver"
        write(*,*) "=================================="
        write(*,*) ""

        N1 = Maxi*NN
        M1 = Maxi*NN

        IT = 10 
        write(*,*) N1, M1, IT

        CALL ITER_ILUT(N1,M1,A,SST,b,IT)

        write(*,*) ""
        write(*,*) "=============================================="
        write(*,*) "The MISSION of Iterative Solver is COMPLETE"
        write(*,*) "=============================================="
        write(*,*) ""

        ELSE
            write(*,*) "Please, Dens or Spare Matrix only !!!"
                STOP    
        ENDIF

        DEALLOCATE ( A )

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

	su=0.0D0
	DO KK=1, NN
		DO I=1, Maxi
                ENG = (I-1)*DELE
			S(KK,I) = SST(I+(KK-1)*Maxi)
			su = su + S(KK,I)
                write(11111,9901) KK, I, ENG, S(KK,I)
		ENDDO
	ENDDO

        write(*,*) "sum of CC(i)_ss = ", su

9901    FORMAT(2(I5,2x),F12.2,2x,ES20.10)

        DEALLOCATE ( SS, DS )

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

        IF(FEDISS.EQ.'YES') THEN  !!! To print out F(E)J in files
!!!                                     fort.701, 702, 703, so on 

        DO KK=1, NDISS

        su2 = 0.0d0
        su3 = 0.0d0
        su4 = 0.0d0
        DO I=1, Maxi
           FF(I) = SST(I+(NDI(KK)-1)*Maxi)*rk(NDI(KK),NDJ(KK),I)
           su2 = su2 + FF(I)

           Fint(I) = SST(I+(NDI(KK)-1)*Maxi)
           su3 = su3 + Fint(I)

           RRKM(I) = rk(NDI(KK),NDJ(KK),I)
 
        ENDDO

        DO I=1, Maxi
                ENG = (I-1)*DELE
                RRKM(I) = RRKM(I)*Fint(I)/su3
                su4 = su4 + RRKM(I)
                FF(I) = FF(I)/su2
                Fint(I) = Fint(I)/su3
 
                write(700+KK,993) I, ENG, FF(I)
                write(800+KK,993) I, ENG, Fint(I)
                write(900+KK,993) I, ENG, RRKM(I)
        ENDDO

        write(2000+KK,9931) su4

        ENDDO
        ENDIF

993     FORMAT(I5,2X,F12.2,2X,ES15.8)
9931    FORMAT(ES15.8)

!CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC


!CCC Computing rates for each product 

1111      IF(suJJ.eq.0.0d0) THEN
                suJ = 1.0d0
          ENDIF      

          write(222,'(I20)') JJT
          write(222,'(ES20.10)') suJJ/suJ

          write(555,'(I20)') JJT
          write(555,'(ES20.10)') suJJ/suJ

          write(666,'(ES20.10)') suJJ/suJ

          write(*,'(A20,2X,I20)') "J = ", JJT
          write(*,'(A20,2X,ES20.10)') "sum of F(E) = ", suJJ/suJ

          DO KK=1, NN
                DO L=1, ip(KK)
                 IF(L.eq.1) THEN
                    Rate(L)=0.0D0
                    DO I=1, Nsink(KK)-1
                      Rate(L)=Rate(L) + S(KK,I)*RKc*rp(KK,L,I)*suk           
                    ENDDO
                 ELSE

                 Rate(L)=0.0D0
                 DO I=Nsink(KK), Maxi
                    Rate(L)=Rate(L) + S(KK,I)*RKc*rp(KK,L,I)*suk
                 ENDDO
                 ENDIF
                 Rate0(KK,L) = Rate0(KK,L) + Rate(L)
                ENDDO
          ENDDO

        IF((suJJ.eq.0.0d0).and.(suJ.eq.1.0d0)) goto 2222

!!!       ENDDO  !!! end of loop J

        DEALLOCATE (rp, rk)

450        write(*,*) ""
        write(*,*) "............................."

	DO I=1, NN
          DO J=1, ip(I)
                IF(Rate0(I,J).LT.0.0d0) Rate0(I,J) = 0.0d0  !!! No NEGATIVE
                IF(Rate0(I,J).GT.RKc) Rate0(I,J) = RKc   !!! No greater than RKc
                write(*,'(A18, A1, ES20.5)')  CNAME(I,J), "=", Rate0(I,J)
                write(222,'(ES20.5)') Rate0(I,J)
          Enddo
          write(*,*) "............................."
!!          write(222,*) "............................."
        Enddo

2222        write(*,*) '' 
        write(*,*) ("=", KK=1, 58)
        write(*,*) " SUMMARY of RESULTS : "
        write(*,*) ("=", KK=1, 58) 
        write(*,*) '' 
        write(*,*) "Thermal rate constants (or yields) for various products:"
        write(*,*) '' 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        DO I=1, NN+Np
                RateP(I)=0.0d0
        ENDDO 
        J=1
        DO I=1, NN
                RateP(I) = RateP(I) + Rate0(I,J)
        ENDDO
        DO J=1, Np
                DO I=1, NN
                        RateP(J+NN) = RateP(J+NN) + Rate0(I,J+1)
                ENDDO
        ENDDO

        DO I=1, NN+Np
                IF(RateP(I).LT.0.0d0) RateP(I)=0.0d0  !!! No NEGATIVE
                IF(RateP(I).GT.RKc) RateP(I)=RKc   !!! No greater than RKc

                IF(I.LE.NN) THEN
        write(444,'(I3, A8, A20, A1, ES15.5)')  I, ") Stab. ", CNM(I), "=", RateP(I)
        write(*,'(I3, A8, A20, A1, ES15.5)')  I, ") Stab. ", CNM(I), "=", RateP(I)
        write(555,'(ES15.5)') RateP(I)
                ELSE
        write(444,'(I3, A8, A20, A1, ES15.5)')  I, ")       ", CNP(I-NN), "=", RateP(I)
        write(*,'(I3, A8, A20, A1, ES15.5)')  I, ")       ", CNP(I-NN), "=", RateP(I)
        write(555,'(ES15.5)') RateP(I)
                ENDIF
        ENDDO

        write(*,*) ''
        write(*,*) ("=", KK=1, 58)
        write(*,*) " END of STHREE PROGRAM "
        write(*,*) ("=", KK=1, 58)
        write(*,*) ''

        STOP
        RETURN
	END
